"""Scanner for PR force-pushed dangling-commits"""

import requests
from rich.console import Console

from dangling_finder.exceptions import GitHubRepoError

err_console = Console(stderr=True)


class _GraphQL:
    """Class for graphQL queries of force-pushed PRs"""

    def __init__(self, owner, repo, github_token):
        self._github_token = github_token
        self._repo = repo
        self._owner = owner
        self._rest_headers = {
            "X-GitHub-Api-Version": "2022-11-28",
            "Authorization": f"Bearer {self._github_token}",
            "Accept": "application/vnd.github+json",
        }

    def check_repository(self):
        url_api = f"https://api.github.com/repos/{self._owner}/{self._repo}"
        url_repo = f"https://github.com/{self._owner}/{self._repo}"
        err_console.print(f"Loading dangling commits on GitHub: {url_repo}")
        r = requests.get(url_api, headers=self._rest_headers, timeout=10)
        if r.status_code != 200:
            raise GitHubRepoError(
                f"Could not connect to the following repo: {url_repo}"
            )
        err_console.print("✅ GitHub repository found")

    def get_pull_request_highest_number(self):
        url = (
            f"https://api.github.com/repos/{self._owner}/{self._repo}/pulls"
            "?state=all"
        )
        resp = requests.get(url, headers=self._rest_headers, timeout=10)
        body = resp.json()
        if len(body) == 0:
            return 0
        return body[0]["number"]

    def extract_from_single_graphql_response(self, result):
        found_dangling_heads = []
        found_closed_not_merged_prs = []
        has_next_page = result["data"]["repository"]["pullRequests"][
            "pageInfo"
        ]["hasNextPage"]
        new_cursor = result["data"]["repository"]["pullRequests"]["pageInfo"][
            "endCursor"
        ]
        cost_add = result["data"]["rateLimit"]["cost"]

        result_data = result["data"]["repository"]["pullRequests"]["nodes"]
        # 1. extract heads of PR not merged and closed
        for pull_request in result_data:
            if (pull_request["closed"] is True) and (
                pull_request["merged"] is False
            ):
                found_closed_not_merged_prs.append(pull_request["id"])
            forced_pushed_events = pull_request["timelineItems"]["nodes"]
            # 2. extract heads from force-pushed events
            if len(forced_pushed_events) > 0:
                for dangling_head in forced_pushed_events:
                    before_commit = dangling_head["beforeCommit"]
                    # see issue 6
                    if before_commit is not None:
                        commit_author = before_commit["author"]
                        commit_head = {
                            "commitUrl": before_commit["commitUrl"],
                            "commitSHA": before_commit["commitUrl"].split("/")[
                                -1
                            ],
                            "author": {
                                "name": commit_author["name"],
                                "email": commit_author["email"],
                            },
                            "date": before_commit["authoredDate"],
                            "type": {
                                "name": "ForcePushed PR",
                                "PR": pull_request["url"],
                            },
                        }
                        found_dangling_heads += [commit_head]
        return (
            has_next_page,
            new_cursor,
            cost_add,
            found_dangling_heads,
            found_closed_not_merged_prs,
        )

    def extract_closed_pr_response(self, result):
        closed_pr_head_commits = []
        cost_add = result["data"]["rateLimit"]["cost"]
        result_data = result["data"]["nodes"]
        for node in result_data:
            # if PR force-pushed to begin (no commit in PR)
            # cf. https://github.com/akeneo/pim-community-dev/pull/2025
            if len(node["commits"]["nodes"]) > 0:
                pull_request = node
                commit = node["commits"]["nodes"][0]["commit"]
                commit_author = commit["author"]
                commit_result = {
                    "commitUrl": commit["commitUrl"],
                    "commitSHA": commit["commitUrl"].split("/")[-1],
                    "author": {
                        "name": commit_author["name"],
                        "email": commit_author["email"],
                    },
                    "date": commit["authoredDate"],
                    "type": {
                        "name": "Closed PR",
                        "PR": pull_request["url"],
                    },
                }
                closed_pr_head_commits.append(commit_result)
        return cost_add, closed_pr_head_commits

    def execute_closed_pr_queries(self, pr_list, previous_rate_limit):
        if len(pr_list) == 0:
            return [], previous_rate_limit
        dangling_heads = []
        total_cost = 0
        query = """
        query ($ids: [ID!]!) {
        rateLimit {
            resetAt
            cost
            remaining
        }
        nodes(ids: $ids) {
            ... on PullRequest {
            url
            commits(last: 1) {
                nodes {
                commit {
                    authoredDate
                    author { name, email }
                    commitUrl
                }
                }
            }
            }
        }
        }
        """
        nb_queries = len(pr_list) // 100 + (len(pr_list) % 100 > 0)
        for i in range(nb_queries):
            variables = {"ids": pr_list[i * 100 : (i + 1) * 100]}
            request = requests.post(
                "https://api.github.com/graphql",
                json={"query": query, "variables": variables},
                headers={
                    "Authorization": f"Bearer {self._github_token}",
                    "X-Github-Next-Global-ID": "1",
                },
                timeout=10,
            )
            if request.status_code == 200:
                result = request.json()
                cost_add, found_dangling_closed_commits = (
                    self.extract_closed_pr_response(result)
                )
                dangling_heads += found_dangling_closed_commits
                total_cost += cost_add
            else:
                raise GitHubRepoError(
                    f"""Query failed to run, code: {request.status_code}.
                    Response body:\n{request.text}
                    Response headers:\n{request.headers}"""
                )
        remaining_rate_limit = result["data"]["rateLimit"]
        remaining_rate_limit["total"] = (
            total_cost + previous_rate_limit["total"]
        )
        return dangling_heads, remaining_rate_limit

    def execute_force_pushed_queries(self):
        i = 0
        end_cursor = ""
        dangling_heads = []
        closed_prs = []
        total_cost = 0
        total_time = 0
        query = """
        query ($owner: String!, $repo: String!, $endCursor: String!) {
        rateLimit {
            resetAt
            cost
            remaining
        }
        repository(name: $repo, owner: $owner ) {
            pullRequests(first: 100, after: $endCursor) {
            pageInfo {
                hasNextPage
                endCursor
            }
            nodes {
                ... on PullRequest {
                id
                url
                number
                closed
                merged
                timelineItems(first: 100, itemTypes: \
        [HEAD_REF_FORCE_PUSHED_EVENT]) {
                    nodes {
                    ... on HeadRefForcePushedEvent {
                        beforeCommit {
                        authoredDate
                        author { name, email }
                        commitUrl
                        }
                    }
                    }
                }
                }
            }
            }
        }
        }
        """
        while True:
            current_query = query
            variables = {
                "owner": self._owner,
                "repo": self._repo,
                "endCursor": end_cursor,
            }
            if total_cost == 0:
                current_query = current_query.replace(
                    ", $endCursor: String!", ""
                )
                current_query = current_query.replace(", after: $endCursor", "")
                variables.pop("endCursor")
            request = requests.post(
                "https://api.github.com/graphql",
                json={"query": current_query, "variables": variables},
                headers={
                    "Authorization": f"Bearer {self._github_token}",
                    "X-Github-Next-Global-ID": "1",
                },
                timeout=10,
            )
            i += 1
            if request.status_code == 200:
                result = request.json()
                (
                    has_next_page,
                    new_cursor,
                    cost_add,
                    found_dangling_heads,
                    found_closed_pr_heads,
                ) = self.extract_from_single_graphql_response(result)
                total_cost += cost_add
                total_time += request.elapsed.total_seconds()
                dangling_heads += found_dangling_heads
                closed_prs += found_closed_pr_heads
                if has_next_page:
                    end_cursor = new_cursor
                else:
                    break
            else:
                raise GitHubRepoError(
                    f"""Query failed to run, code: {request.status_code}.
                    Response body:\n{request.text}
                    Response headers:\n{request.headers}"""
                )
        remaining_rate_limit = result["data"]["rateLimit"]
        remaining_rate_limit["total"] = total_cost
        remaining_rate_limit["total_time"] = total_time
        return (
            dangling_heads,
            remaining_rate_limit,
            closed_prs,
        )



"""CLI module of dangling-finder"""

import time
import typer
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn
from typing_extensions import Annotated

from dangling_finder.listing import _GraphQL
from dangling_finder.output import OutputFormat

err_console = Console(stderr=True)

app = typer.Typer(
    context_settings={"help_option_names": ["-h", "--help"]},
    no_args_is_help=True,
)


def mutually_exclusive_group(size=1):
    group = []

    def callback(
        ctx: typer.Context, param: typer.CallbackParam, value: bool
    ):  # pylint: disable=unused-argument
        current_param = param.name.replace("_", "-")
        if value is True and current_param not in group:
            group.append(current_param)
        if len(group) > size:
            raise typer.BadParameter(
                f"{current_param} is mutually exclusive with {group[-2]}"
            )
        return value

    return callback


exclusivity_callback = mutually_exclusive_group()


@app.command("pull-requests", no_args_is_help=True)
def find_lost_pr_heads(
    owner: str,
    repo: str,
    github_token: Annotated[str, typer.Argument(envvar="GITHUB_TOKEN")],
    bash_script: Annotated[
        bool, typer.Option(callback=exclusivity_callback)
    ] = False,
    git_config: Annotated[
        bool, typer.Option(callback=exclusivity_callback)
    ] = False,
    batch: Annotated[int, typer.Option()] = 0,
):
    """List dangling commits SHA-1 in a GitHub repository's pull requests.
    NB: Only upper parents are returned.

    Args:
        owner (str): name of the repository owner
        repo (str): name of the repository
        github_token (str): personnal GitHub access token
        bash_script (bool): return bash script for local git repo
        git_config (bool): return a confgi config text to append
        batch (int): batch size of JSON output only.
    """
    graphql_api = _GraphQL(owner, repo, github_token)
    graphql_api.check_repository()
    pr_max = graphql_api.get_pull_request_highest_number()
    err_console.print(f"{pr_max} pull requests to scan.")
    start_time = time.time()
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
        console=err_console,
    ) as progress:
        task1 = progress.add_task(
            description="Get all force-pushed events in PRs...", total=10
        )
        result1, old_rate_limit, prs = (
            graphql_api.execute_force_pushed_queries()
        )
        progress.update(
            task1, completed=10, description="Force-pushed PR events retrieved."
        )
        task2 = progress.add_task(
            description="Get all closed-and-not-merged PRs...", total=10
        )
        result2, rate_limit = graphql_api.execute_closed_pr_queries(
            prs, old_rate_limit
        )
        progress.update(
            task2,
            completed=10,
            description="Closed-and-not-merged PRs retrieved.",
        )
    duration = time.time() - start_time
    err_console.print(
        "Done. Duration: " + time.strftime("%H:%M:%S", time.gmtime(duration))
    )
    err_console.print("GitHub API quotas:")
    err_console.print(f'Remaining rate limit - {rate_limit["remaining"]}')
    err_console.print(f'Reset date rate limit - {rate_limit["resetAt"]}')
    err_console.print(f'Total cost used - {rate_limit["total"]}')
    output = OutputFormat(
        owner=owner,
        repo=repo,
        bash_script=bash_script,
        git_config=git_config,
        json_dangling_heads=result1 + result2,
        batch=batch
    ).output()
    typer.echo(output)


@app.command("events", no_args_is_help=True)
def find_latest_dangling_push_events():
    """WIP: find some danglign commits using latest push events
    """


if __name__ == "__main__":
    app()



"""Module used for output generation"""

import json

from dataclasses import dataclass
from textwrap import dedent


@dataclass
class OutputFormat:
    """Class for generating output with correct formatting"""

    owner: str
    repo: str
    bash_script: bool
    git_config: bool
    json_dangling_heads: list
    batch: int

    @property
    def dangling_heads(self):
        dangling_heads = [
            e["commitUrl"][::-1].split("/", 1)[0][::-1]
            for e in self.json_dangling_heads
        ]
        return dangling_heads

    def generate_bash_script(self):
        dangling_heads = self.dangling_heads
        regroup_git_commands = []
        cmdline_kernel_length_limit = 4096
        current_command = (
            f"git fetch origin {dangling_heads[0]}"
            f":refs/remotes/origin/dangling-{dangling_heads[0][:10]}"
        )
        next_command = (
            f"git fetch origin {dangling_heads[0]}"
            f":refs/remotes/origin/dangling-{dangling_heads[0][:10]}"
        )
        i = 1
        while i < len(dangling_heads):
            while len(next_command) < cmdline_kernel_length_limit and i < len(
                dangling_heads
            ):
                current_command = next_command
                next_command = (
                    current_command
                    + f" {dangling_heads[i]}"
                    + f":refs/remotes/origin/dangling-{dangling_heads[i][:10]}"
                )
                i += 1
            if len(next_command) < cmdline_kernel_length_limit:
                continue
            else:
                regroup_git_commands.append(current_command)
                next_command = (
                    f"git fetch origin {dangling_heads[i-1]}"
                    f":refs/remotes/origin/dangling-{dangling_heads[i-1][:10]}"
                )
        regroup_git_commands.append(next_command)
        return "\n".join(regroup_git_commands)

    def generate_git_config(self):
        dangling_heads = self.dangling_heads
        template = f"""
        [remote "dangling-REPLACE"]
        \turl = git@github.com:{self.owner}/{self.repo}.git
        \tfetch = REPLACE:refs/remotes/origin/dangling-REPLACE"""
        template = dedent(template)
        git_config = [
            template.replace("REPLACE", hash) for hash in dangling_heads
        ]
        return "\n".join(git_config)

    def output(self):
        if self.bash_script:
            return self.generate_bash_script()
        elif self.git_config:
            return self.generate_git_config()

        if self.batch > 0:
            out_array = [
                self.json_dangling_heads[i : i + self.batch]
                for i in range(0, len(self.json_dangling_heads), self.batch)
            ]
        else:
            out_array = self.json_dangling_heads

        return json.dumps(out_array, indent=4)
